"use strict";

var Node = require("./Node");

function shadowCopy(src, target) {
  for (var k in src) {
    if (src.hasOwnProperty(k)) {
      target[k] = src[k];
    }
  }
}

function newInstance() {
  var meta = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];

  var m = {};

  m._meta = meta;
  m._indexers = new Map();

  m.Node = Node; // Node class
  m.Bouton = Node; // alias of node

  m.END = Node.END; // END signal

  m.reserved = ["id", "options", "ee", "observers", "upstreams", "downstreams", "meta", "push", "onReceive", "onSignal", "onError", "onEnd", "send", "observe", "to", "pull", "onRequest", "request", "from", "isErrorSignal", "isEndSignal", "throwError", "invokeObservers"];

  function indexNode(node) {
    var _iteratorNormalCompletion = true;
    var _didIteratorError = false;
    var _iteratorError = undefined;

    try {
      for (var _iterator = m._indexers.values()[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
        var indexer = _step.value;

        indexer(node);
      }
    } catch (err) {
      _didIteratorError = true;
      _iteratorError = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion && _iterator.return) {
          _iterator.return();
        }
      } finally {
        if (_didIteratorError) {
          throw _iteratorError;
        }
      }
    }
  }

  /**
   * Add an operator
   * @param  {string} name   - the name of the operator
   * @param  {function} operator - the operator function
   * @return {bouton}          the bouton module
   */
  function addOperator(name, operator) {
    if (m.reserved.indexOf(name) >= 0) {
      console.warn("can't add operator '" + name + "', name is reserved. ");
      return m;
    }

    function fn() {
      var node = operator.apply(undefined, arguments);
      node.meta = {};
      shadowCopy(this.meta, node.meta);
      indexNode(node);
      return this.to(node);
    };

    Node.prototype[name] = fn;
    m[name] = function () {
      var node = operator.apply(undefined, arguments);
      node.meta = {};
      shadowCopy(m._meta, node.meta);
      indexNode(node);
      return node;
    };

    return m;
  }
  m.addOperator = addOperator;

  /**
   * Add multiple operators
   * @param {object} operators - name:operator pair
   * @return {bouton} the bouton module
   */
  function addOperators(ops) {
    for (var name in ops) {
      m.addOperator(name, ops[name]);
    }
  }
  m.addOperators = addOperators;

  /**
   * add a source
   * @param {string} name   - the source name
   * @param {function} source - the source function
   * @return {bouton}  the bouton module
   */
  function addSource(name, source) {
    if (m.reserved.indexOf(name) >= 0) {
      console.warn("can't add source '" + name + "', name is reserved. ");
      return m;
    }
    m[name] = function () {
      var node = source.apply(undefined, arguments);
      node.meta = {};
      shadowCopy(m._meta, node.meta);
      indexNode(node);
      return node;
    };
    return m;
  }
  m.addSource = addSource;

  /**
   * Add multiple sources
   * @param {object} srcs - name:source pair
   * @return {bouton} the bouton module
   */
  function addSources(srcs) {
    for (var name in srcs) {
      m.addSource(name, srcs[name]);
    }
    return m;
  }
  m.addSources = addSources;

  /**
   * add querier 
   * @param {string} name - the querier name
   * @param {object} querier - the querier object with query method
   * @return {bouton} the bouton module
   */
  function addQuerier(name, querier) {
    m[name] = function () {
      for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
        args[_key] = arguments[_key];
      }

      return querier.query.apply(querier, [m].concat(args));
    };
    return m;
  }
  m.addQuerier = addQuerier;

  /**
   * add indexer 
   * @param {string} name - the indexer name
   * @param {object} indexer - the indexer object with index method
   * @return {bouton} the bouton module
   */
  function addIndexer(name, indexer) {
    m._indexers.set(name, function (node) {
      indexer.index(m, node);
    });
    return m;
  }
  m.addIndexer = addIndexer;

  /**
   * load default sources and operators
   * @return {[type]} [description]
   */
  function extendDefault() {
    var operators = require("./operators"); // default operators
    var sources = require("./sources"); // default sources

    m.addSources(sources);
    m.addOperators(operators);
    return m;
  }
  m.default = extendDefault;

  m.observers = {};

  function extend(extension) {
    if (typeof extension === "string") {
      extension = require(extension);
    }

    if (extension.operators) {
      m.addOperators(extension.operators);
    }

    if (extension.sources) {
      m.addSources(extension.sources);
    }

    if (extension.observers) {
      for (var name in extension.observers) {
        m.observers[name] = extension.observers[name];
      }
    }

    if (extension.indexers) {
      for (var _name in extension.indexers) {
        m.addIndexer(_name, extension.indexers[_name]);
      }
    }

    if (extension.queriers) {
      for (var _name2 in extension.queriers) {
        m.addQuerier(_name2, extension.queriers[_name2]);
      }
    }

    if (extension.others) {
      for (var _name3 in extension.others) {
        m[_name3] = extension.others[_name3];
      }
    }

    return m;
  }
  m.extend = extend;

  /**
   * set meta for the bouton object
   */
  function setMeta(metadata) {
    m._meta = metadata;
  }
  m.setMeta = setMeta;

  /**
   * set single meta data for the bouton object
   */
  function addMeta(name, meta) {
    m._meta[name] = meta;
  }
  m.addMeta = addMeta;

  /**
   * remove meta data
   */
  function removeMeta(name) {
    if (m._meta.hasOwnProperty(name)) {
      delete m._meta[name];
    }
  }
  m.removeMeta = removeMeta;

  return m;
}

var m = newInstance();
m.new = newInstance;

module.exports = m;